#!/usr/bin/Rscript

# https://cran.r-project.org/web/packages/segclust2d/vignettes/segclust.html
# cannot be install through conda
# created a conda env. w/ R
# install.packages("segclust2d")

require(ggplot2)
require(scales)
require(segclust2d)

############################## FUNC

cov_seg <- function(x, lmin=1e4){
    set.seed(42)
    suppressMessages(
        x_seg <- segclust2d::segmentation(
            data.frame(
                base=seq(1, length(x)),
                cov=x
            ),
            lmin=lmin,
            Kmax=5,
            seg.var=c("base", "cov"),
            scale.variable=FALSE,
            subsample=TRUE,
            subsample_over=1000
        )
    )
    return(x_seg)
}

cov_seg_sum <- function(x, lmin=1e4){
    x_length <- length(x)
    x_mean   <- mean(x)
    x_median <- median(x)
    x_sd     <- sd(x)
    if(length(x) < 2*lmin){
        # stats
        x_states        <- 0
        x_medians       <- x_median
        x_medians_sd    <- NA
    } else {
        # segmentation
        x_seg           <- cov_seg(x=x, lmin=lmin)
        x_aug           <- segclust2d::augment(x_seg)
        # stats
        x_states        <- nrow(segclust2d::states(x_seg))
        x_medians       <- aggregate(x=x_aug$cov, by=list(state=x_aug$state), FUN=median)$x
        x_medians_sd    <- sd(x_medians)
    }
    return(c(
        length=x_length,
        mean=x_mean,
        median=x_median,
        sd=x_sd,
        states=x_states,
        states_medians=paste(x_medians, collapse=";"),
        states_medians_sd=x_medians_sd
    ))
}

cov_seg_plot <- function(df, cid, lmin=1e4){
    df <- df[df$contig==cid,]
    df_seg <- augment(cov_seg(x=df$cov, lmin=lmin))
    df_seg$state_median <- NA
    for(s in unique(df_seg$state)){
        df_seg$state_median[df_seg$state == s] <- median(df_seg[df_seg$state==s, "cov"])
    }
    iplot <-
        ggplot(data=df_seg, aes(x=base, y=cov)) +
        geom_line(colour="#666666") +
        geom_line(aes(x=base, y=state_median), colour="#0066CC", size=2) +
        scale_x_continuous(breaks=seq(0, max(df_seg$base), by=5000)) +
        scale_y_log10(breaks=trans_breaks("log10", function(x) 10^x), labels=trans_format("log10", math_format(10^.x))) +
        labs(
            title=cid,
            x="base",
            y="coverage (metaG, SR)"
        ) +
        theme_bw() +
        theme(
            axis.text.x=element_text(angle=90, vjust=0.5, hjust=1)
        )
    return(iplot)
}

############################## DATA
LMIN <- 1e+4

df <- read.csv(
    file="/scratch/users/vgalata/GDB/results/mapping/metag/lr/flye_sr/ASSEMBLY.FILTERED.sr.cov.perbase",
    sep="\t", header=FALSE, col.names=c("contig", "base", "cov"), stringsAsFactors=FALSE
)

df_states  <- aggregate(x=df$cov, by=list(contig=df$contig), FUN=cov_seg_sum, lmin=LMIN, simplify=TRUE)
df_states  <- data.frame(
    contig              = df_states$contig,
    length              = as.integer(df_states$x[,"length"]),
    mean                = as.numeric(df_states$x[,"mean"]),
    median              = as.numeric(df_states$x[,"median"]),
    sd                  = as.numeric(df_states$x[,"sd"]),
    states              = as.integer(df_states$x[,"states"]),
    states_medians      = df_states$x[,"states_medians"],
    states_medians_sd   = as.numeric(df_states$x[,"states_medians_sd"]),
    stringsAsFactors=FALSE
)

# scatter plot
pscatter <-
    ggplot(data=df_states[df_states$states > 1,], aes(x=length, y=mean)) +
    geom_point(data=df_states[df_states$states < 2,], aes(x=length, y=mean), shape=4, color="#CCCCCC", size=0.5) +
    geom_point(aes(size=sd, fill=as.factor(states), color=as.factor(states)), shape=21, alpha=0.75) +
    scale_x_log10(
        breaks=trans_breaks("log10", function(x) 10^x),
        labels=trans_format("log10", math_format(10^.x))
    ) +
    scale_y_log10(
        breaks=trans_breaks("log10", function(x) 10^x),
        labels=trans_format("log10", math_format(10^.x))
    ) +
    labs(
        title="Contig coverage segmentation",
        subtitle="Flye + Racon(metaG SR)",
        fill="Number of states",
        color="Number of states",
        size="State median SD",
        x="Contig length [bp]",
        y="Mean contig coverage (metaG, SR)"
    ) +
    theme_bw()
pdf("tmp/covseg2.pdf", width=7, height=7); print(pscatter); dev.off()

# contig coverage: (random) contigs w/ more than one state and relat. high state median SD
pdf("tmp/covseg.pdf", width=20, height=4)
set.seed(42)
rcids <- df_states[df_states$states>1 & df_states$states_medians_sd>=100,"contig"]
rcids <- sort(sample(rcids, min(10, length(rcids))))
for(cid in rcids){ print(cov_seg_plot(df, cid, LMIN)) }
dev.off()



# # scatter plot: number of states vs state median SD (if more than one state)
# df_medians <- data.frame(
#     contig=rep(df_states$contig, times=sapply(df_states$states, function(x){ ifelse(x==0,1,x) })),
#     median=NA,
#     state=NA,
#     stringsAsFactors=FALSE
# )
# j <- 1
# for(i in 1:nrow(df_states)){
#     i_n <- max(1, df_states[i, "states"])
#     df_medians[j:(j+i_n-1), "contig"] <- df_states[i, "contig"]
#     df_medians[j:(j+i_n-1), "state"]  <- 1:i_n
#     df_medians[j:(j+i_n-1), "median"] <- as.numeric(unlist(strsplit(df_states[i, "medians"], ";")))
#     j <- j + i_n
# }
# pscatter <-
#     ggplot(data=df_medians, aes(x=contig, y=median, size=state)) +
#     geom_point(shape=21,colour="white", fill="#333333") +
#     scale_y_log10(
#         breaks=trans_breaks("log10", function(x) 10^x),
#         labels=trans_format("log10", math_format(10^.x))
#     ) +
#     theme_bw() +
#     theme(
#         axis.text.x=element_blank(),
#         axis.ticks.x=element_blank(),
#         panel.grid.major = element_blank(),
#         panel.grid.minor = element_blank()
#     )
# pdf("tmp/covseg2.pdf", width=20, height=4); print(pscatter); dev.off()
