#!/usr/bin/Rscript

## IMPORT
suppressMessages(library(ggsci)) # colors

##############################
# INPUT
read_nanostats <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    df_cols <- c("stat"="Statistic", "value"="Value")
    colnames(df) <- df_cols[colnames(df)]
    return(df)
}

read_quast <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        row.names=1,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    testit::assert(all(colnames(df) %in% names(ASM_TOOL_NAMES)))
    colnames(df) <- ASM_TOOL_NAMES[colnames(df)]
    df <- df[QUAST_VARS, ASM_TOOL_NAMES]
    return(df)
}

read_plasflow <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        stringsAsFactors=FALSE,
        check.names=FALSE
    )
    testit::assert(all(df$tool %in% names(ASM_TOOL_NAMES)))
    df$tool <- ASM_TOOL_NAMES[df$tool]
    df <- df[df$label %in% names(PLASFLOW_NAMES$labels),]
    df$label <- PLASFLOW_NAMES$labels[df$label]
    return(df)
}

read_rgi <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        stringsAsFactors=FALSE,
        check.names=FALSE
    )
    colnames(df) <- sapply(colnames(df), function(x){ ifelse(x %in% names(ASM_TOOL_NAMES), ASM_TOOL_NAMES[x], x) })
    testit::assert(all(df$col %in% names(RGI_NAMES$col)))
    df$col <- RGI_NAMES$col[df$col]
    return(df)
}

read_barrnap <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        stringsAsFactors=FALSE,
        check.names=FALSE
    )
    testit::assert(all(df$tool %in% names(ASM_TOOL_NAMES)))
    df$tool <- ASM_TOOL_NAMES[df$tool]
    testit::assert(all(df$kingdom %in% names(BARRNAP_KINGDOM_NAMES)))
    df$kingdom <- BARRNAP_KINGDOM_NAMES[df$kingdom]
    return(df)
}

read_crispr <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        stringsAsFactors=FALSE,
        check.names=FALSE
    )
    testit::assert(all(df$crispr_tool %in% names(CRISPR_TOOL_NAMES)))
    testit::assert(all(df$asm_tool    %in% names(ASM_TOOL_NAMES)))
    df$crispr_tool <- CRISPR_TOOL_NAMES[df$crispr_tool]
    df$asm_tool    <- ASM_TOOL_NAMES[df$asm_tool]
    return(df)
}

read_mappability <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    testit::assert(all(df$tool %in% names(ASM_TOOL_NAMES)))
    df$tool <- ASM_TOOL_NAMES[df$tool]
    return(df)
}

read_diamond <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    testit::assert(all(df$tool %in% names(ASM_TOOL_NAMES)))
    df$tool <- ASM_TOOL_NAMES[df$tool]
    return(df)
}

read_prodigal <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    testit::assert(all(df$tool %in% names(ASM_TOOL_NAMES)))
    df$tool <- ASM_TOOL_NAMES[df$tool]
    return(df)
}

read_prodigal_gcounts <- function(fname){
    df <- read_prodigal(fname)
    df$partial_pct <- 100 * df$partial / df$total
    return(df)
}

read_prodigal_glength <- function(fname){
    df <- read_prodigal(fname)
    return(df)
}

read_cdhit <- function(fname){
    print(sprintf("Reading: %s", fname))
    df <- read.csv(
        file=fname,
        sep="\t",
        header=TRUE,
        check.names=FALSE,
        stringsAsFactors=FALSE
    )
    df <- df[,c("tool1", "tool2", "mtype", "total", "unique", "highcov", "highcovuniq")]
    df$unique_pct               <- 100 * df$unique / df$total
    df$highcovunique_pct_total  <- 100 * df$highcovuniq / df$total
    df$highcovunique_pct_unique <- 100 * df$highcovuniq / df$unique
    testit::assert(all(df$tool1 %in% names(ASM_TOOL_NAMES)))
    testit::assert(all(df$tool2 %in% names(ASM_TOOL_NAMES)))
    testit::assert(all(df$mtype %in% names(META_TYPES)))
    df$tool1 <- ASM_TOOL_NAMES[df$tool1]
    df$tool2 <- ASM_TOOL_NAMES[df$tool2]
    df$mtype <- META_TYPES[df$mtype]
    return(df)
}

read_mash_dist_reads <- function(fname){
    proc_name <- function(x){
        rtype <- basename(dirname(x))
        mtype <- basename(dirname(dirname(x)))
        testit::assert(rtype %in% names(READ_TYPES))
        testit::assert(mtype %in% names(META_TYPES))
        return(sprintf("%s %s", META_TYPES[mtype], READ_TYPES[rtype]))
    }
    dm <- read.csv(file=fname, sep='\t', header=TRUE, row.names=1, check.names=FALSE)
    colnames(dm) <- sapply(colnames(dm), proc_name)
    rownames(dm) <- sapply(rownames(dm), proc_name)
    return(dm)
}

read_mash_dist_asm <- function(fname){
    proc_name <- function(x){
        x <- basename(dirname(x))
        testit::assert(x %in% names(ASM_TOOL_NAMES))
        return(ASM_TOOL_NAMES[x])
    }
    dm <- read.csv(file=fname, sep='\t', header=TRUE, row.names=1, check.names=FALSE)
    colnames(dm) <- sapply(colnames(dm), proc_name)
    rownames(dm) <- sapply(rownames(dm), proc_name)
    # testit::assert(all(colnames(dm) %in% names(ASM_TOOL_NAMES)))
    # colnames(dm) <- ASM_TOOL_NAMES[colnames(dm)]
    # rownames(dm) <- ASM_TOOL_NAMES[rownames(dm)]
    return(dm)
}

##############################
# PLOTS

plot_rgi_overlap <- function(df, ctype, col){
    df <- df[df$type == ctype & df$col == col,]
    df_list <- lapply(ASM_TOOL_NAMES, function(x){ df[df[,x] > 0,"label"] })
    names(df_list) <- ASM_TOOL_NAMES[names(df_list)]
    UpSetR::upset(
        data=UpSetR::fromList(df_list),
        # overlap order
        order.by="degree",
        decreasing=FALSE,
        # number of sets to plot
        nsets=length(ASM_TOOL_NAMES),
        # y-label title
        mainbar.y.label=sprintf("Intersection size (%s hits, %s)", ctype, col),
        # text size
        text.scale = c(1.2, 1.2, 1.2, 1.2, 1.2, 1.2)#,
        # colors
        # set.metadata=list(
        #     data=data.frame(
        #         sets=names(df_list),
        #         Tool=names(df_list),
        #         stringsAsFactors=FALSE
        #     ), 
        #     plots=list(list(type="matrix_rows", column="Tool", colors=ASM_TOOL_COLORS, alpha=0.7))
        # )
    )
}

plot_crispr_overlap <- function(df, asm_tool){
    asm_sets <- lapply(
        CRISPR_TOOL_NAMES,
        function(x){ unique(unlist(df[df$asm_tool==asm_tool & df$crispr_tool == x,"seq_id"])) }
    )
    names(asm_sets) <- CRISPR_TOOL_NAMES[names(asm_sets)]
    print(asm_sets)
    UpSetR::upset(
        data=UpSetR::fromList(asm_sets),
        order.by="degree",
        decreasing=FALSE,
        mainbar.y.label=sprintf("Contig intersection size (%s)", asm_tool),
        sets.x.label="Contigs w/ CRISPR array(s)"
    )
}

plot_cdhit_barplots <- function(df, mtype, ycol, ycol_text){
    pp <-
        ggplot(data=df[df$mtype == mtype,], aes_string(x="tool2", y=ycol, fill="tool2")) +
        geom_col() +
        facet_wrap(vars(tool1), ncol=1) +
        scale_fill_manual(values=ASM_TOOL_COLORS, guide=NULL) +
        labs(
            x="Assembly 2",
            y=sprintf("%s (%s)\n Unique = in Assembly 2 but not in Assembly 1", ycol_text, mtype)
        ) +
        cdhit_theme
    return(pp)
}

plot_cdhit_scatterplot <- function(df, mtype){
    pp <-
        ggplot(data=df[df$mtype == mtype,], aes(x=unique_pct, y=highcovunique_pct_unique, fill=tool2, shape=tool2)) +
        geom_point(colour="white", size=6) +
        scale_fill_manual(values=ASM_TOOL_COLORS, name="") +
        scale_shape_manual(values=ASM_TOOL_SHAPES, name="") +
        labs(
            x="Unique proteins [%, total]",
            y=sprintf("Unique proteins w/ high mean cov. [%%, unique] (%s)", mtype)
        ) +
        default_theme
    return(pp)
}

plot_diamond_density <- function(df, col){
    pp <-
        ggplot(data=df, aes_string(x=col, colour="tool", fill="tool")) +
        geom_density(alpha=0.2) +
        scale_colour_manual(values=ASM_TOOL_COLORS, guide=NULL) +
        scale_fill_manual(values=ASM_TOOL_COLORS, guide=NULL) +
        facet_wrap(vars(tool), ncol=1, scales="free_y") +
        labs(
            x=DIAMOND_VAR_LABELLER(col),
            y="Density"
        ) +
        diamond_theme
    return(pp)
}

# THEMES
default_theme <-
    # theme_bw() +
    theme_minimal(
        base_size=12
    ) +
    theme(
        # plot.title=element_text(size=14, face="bold"),
        # plot.subtitle=element_text(size=12, face="bold"),
        # legend
        legend.title=element_blank(),
        # legend.text=element_text(size=12),
        # grid
        panel.grid=element_blank(),
        # strip
        strip.background=element_rect(fill="white"),
        # strip.text=element_text(size=12, color="black"),
        # axes
        # axis.title=element_text(size=12, color="black"),
        axis.text.y=element_text(size=10, color="black"),
        axis.text.x=element_text(size=10, color="black")
    )

default_theme_axis_text_x <-
    theme(
        axis.text.x=element_text(size=10, color="black", angle=90, vjust=0.5, hjust=1)
    )

mappability_theme <- default_theme + default_theme_axis_text_x

crispr_theme <- default_theme + default_theme_axis_text_x

plasflow_theme <- default_theme + default_theme_axis_text_x

prodigal_theme <- default_theme + default_theme_axis_text_x

diamond_theme <- default_theme + default_theme_axis_text_x

rgi_theme <- default_theme + default_theme_axis_text_x

barrnap_theme <- default_theme + default_theme_axis_text_x

quast_theme <- default_theme + default_theme_axis_text_x

cdhit_theme <- default_theme + default_theme_axis_text_x

##############################
# CONST

###############
# Reads
META_TYPES <- c(
    "metag"="metaG",
    "metat"="metaT"
)
READ_TYPES <- c(
    "sr"="SR",
    "lr"="LR"
)

###############
# Diamond
DIAMOND_VARS <- c(
    "qs_len_ratio"="Query/subject length ratio",
    "qcov"="Query coverage",
    "scov"="Subject coverage"
)
DIAMOND_VAR_LABELLER <- function(x){
    if(x %in% names(DIAMOND_VARS)){
        return(DIAMOND_VARS[x])
    } else {
        return(x)
    }
}

###############
# Assemblers

# names
ASM_TOOL_NAMES <- c(
    "flye"="Flye",
    "wtdbg2"="wtdbg2",
    "canu"="Canu",
    "flye_sr"="Flye + racon(SR)",
    "wtdbg2_sr"="wtdbg2 + racon(SR)",
    "canu_sr"="Canu + racon(SR)",
    "megahit"="MEGAHIT",
    "metaspades"="metaSPAdes",
    "operams"="OPERA-MS",
    "metaspadeshybrid"="metaSPAdes (H)",
    "operams_sr"="OPERA-MS + racon(SR)",
    "metaspadeshybrid_sr"="metaSPAdes (H) + racon(SR)",
    "imp3"="IMP3 (MG/MT)"
)
# select assemblers from config
ASM_TOOL_NAMES <- ASM_TOOL_NAMES[c(
    snakemake@config$assemblers$lr,
    snakemake@config$assemblers$sr,
    snakemake@config$assemblers$hy,
    snakemake@config$assemblers$hyhy
)]
# colors
ASM_TOOL_PALETTE1 <- ggsci::pal_nejm("default", alpha=1)(8)
ASM_TOOL_PALETTE2 <- ggsci::pal_nejm("default", alpha=0.7)(8)
# ASM_TOOL_PALETTE3 <- ggsci::pal_nejm("default", alpha=0.6)(8)
# ASM_TOOL_PALETTE4 <- ggsci::pal_nejm("default", alpha=0.4)(8)
ASM_TOOL_COLORS <- c(
    "Flye"=ASM_TOOL_PALETTE2[1],
    "wtdbg2"=ASM_TOOL_PALETTE2[2],
    "Canu"=ASM_TOOL_PALETTE2[3],
    "Flye + racon(SR)"=ASM_TOOL_PALETTE1[1],
    "wtdbg2 + racon(SR)"=ASM_TOOL_PALETTE1[2],
    "Canu + racon(SR)"=ASM_TOOL_PALETTE1[3],
    "MEGAHIT"=ASM_TOOL_PALETTE1[4],
    "metaSPAdes"=ASM_TOOL_PALETTE1[5],
    "OPERA-MS"=ASM_TOOL_PALETTE2[6],
    "metaSPAdes (H)"=ASM_TOOL_PALETTE2[7],
    "OPERA-MS + racon(SR)"=ASM_TOOL_PALETTE1[6],
    "metaSPAdes (H) + racon(SR)"=ASM_TOOL_PALETTE1[7],
    "IMP3 (MG/MT)"=ASM_TOOL_PALETTE1[8]
)
ASM_TOOL_COLORS <- ASM_TOOL_COLORS[ASM_TOOL_NAMES]

ASM_TOOL_SHAPES <- c(
    "Flye"=22,
    "wtdbg2"=22,
    "Canu"=22,
    "Flye + racon(SR)"=21,
    "wtdbg2 + racon(SR)"=21,
    "Canu + racon(SR)"=21,
    "MEGAHIT"=21,
    "metaSPAdes"=21,
    "OPERA-MS"=22,
    "metaSPAdes (H)"=22,
    "OPERA-MS + racon(SR)"=21,
    "metaSPAdes (H) + racon(SR)"=21,
    "IMP3 (MG/MT)"=21
)
ASM_TOOL_SHAPES <- ASM_TOOL_SHAPES[ASM_TOOL_NAMES]

###############
# CRISPR tools

CRISPR_TOOL_NAMES <- c(
    "minced"="MinCED",
    "casc"="CasC"
)

###############
# PlasFlow

# names
PLASFLOW_NAMES <- list(
    statstype=c(
        count="Sequence count",
        sum="Cumulative sequence length [bp]",
        count_pct="Sequence count [%]",
        sum_pct="Cumulative sequence length [%]"
    ),
    labels=c(
        chromosome="Chromosome",
        plasmid="Plasmid",
        unclassified="Unclassified"
    )
)
# colors
PLASFLOW_COLORS <- list(
    labels=ggsci::pal_nejm("default", alpha=1)(4)[c(2,3,4)]
)
names(PLASFLOW_COLORS$labels) <- PLASFLOW_NAMES$labels

###############
# RGI
RGI_NAMES <- list(
    col=c(
        "Best_Hit_ARO"="ARO term",
        "ARO"="ARO",
        "Drug Class"="Drug class",       
        "Resistance Mechanism"="Resistance mechanism",
        "AMR Gene Family"="Gene family"
    )
)

###############
# barrnap
BARRNAP_KINGDOM_NAMES <- c(
    "arc"="Archea",
    "bac"="Bacteria",
    "euk"="Eukaryota",
    "mito"="Metazoan mitochondria"
)

###############
# QUAST
QUAST_VARS <- c(
    "# contigs",
    "Largest contig",            
    "Total length",
    "N50",
    "L50",                       
    "# N's per 100 kbp"
)